#include "sound.h"
#include "string.h"
#include "stdlib.h"

#include "loadsw32.h"
#include "sw32dsp.h"
#include "mydebug.h"
#include <limits.h>

//  Helper macros
//
#define NUM_ELEMENTS(array)     (sizeof(array) / sizeof(array[0]))

//  Global variables
//
WORD awDmaSpecs[] =  { DMA0, DMA1, DMA3, DMA5, DMA6, DMA7 };
WORD awValidDmas[] = { 0, 1, 3, 5, 6, 7 };
WORD awIrqSpecs[] =  { IRQ3, IRQ5, IRQ7, IRQ9, IRQ10, IRQ11, IRQ12 };
WORD awValidIrqs[] = { 3, 5, 7, 9, 10, 11, 12 };
LONG lNumInPorts = 0;
LONG lNumOutPorts = 0;
INT nPseudoFilePos = 0;

#if FALSE
PSOUND_HARDWARE pHw = NULL;

VOID SetSw32HardwareGlobal (PSOUND_HARDWARE pHardware)
{
    pHw = pHardware;
}

PSOUND_HARDWARE GetSw32HardwareGlobal (VOID)
{
    return (pHw);
}
#endif



WORD InPortWord (UINT wAddress)
{
    // TODO: WORD wResult = INPORT(GetSw32HardwareGlobal(), wAddress);
	// NOTE: Assumes the logical and physical port address are the same
    WORD wResult = (WORD)READ_PORT_USHORT((PUSHORT)wAddress);
	ASSERT(wAddress <= USHRT_MAX);
    lNumInPorts++;

    return (wResult);       
}


VOID OutPortWord (UINT wAddress, UINT wValue)
{
    // TODO: OUTPORT(GetSw32HardwareGlobal(), wAddress, wValue);
	// NOTE: Assumes the logical and physical port address are the same
	ASSERT(wAddress <= USHRT_MAX);
	ASSERT(wValue <= USHRT_MAX);
    WRITE_PORT_USHORT((PUSHORT)wAddress, (USHORT)wValue);
    lNumOutPorts++;
}


//  Pseudo-file operations
//

FILE* OpenBinaryFile (PCSZ pszFile)
{
    nPseudoFilePos = 0;
    return (0);
}

BYTE GetNextByte (FILE* fIn)
{
    //return (bDebugLoad[nPseudoFilePos++]);
    return (abGenmidLoad[nPseudoFilePos++]);
}


BOOL EndOfFile (FILE* fIn)
{
    //return (nPseudoFilePos > nMAX_POS);
    return (nPseudoFilePos > nGenmidLoadSize);
}

VOID CloseBinaryFile (FILE* fIn)
{
}


WORD GetDmaSpec (ULONG uDma)
{
    WORD wDmaSpec = (WORD)-1;
    INT i;
    ASSERT(NUM_ELEMENTS(awValidDmas) == NUM_ELEMENTS(awDmaSpecs));

    for (i = 0; i < NUM_ELEMENTS(awValidDmas); i++) {
        if (awValidDmas[i] == (WORD)uDma) {
            wDmaSpec = awDmaSpecs[i];
            break;
            }
        }
    
    return (wDmaSpec);  
}


WORD GetIrqSpec (ULONG uIrq)
{
    WORD wIrqSpec = (WORD)-1;
    INT i;
    ASSERT(NUM_ELEMENTS(awValidIrqs) == NUM_ELEMENTS(awIrqSpecs));

    for (i = 0; i < NUM_ELEMENTS(awValidIrqs); i++) {
        if (awValidIrqs[i] == (WORD)uIrq) {
            wIrqSpec = awIrqSpecs[i];
            break;
            }
        }
    
    return (wIrqSpec);  
}


//	Initialize the SoundWave32 DSP for emulating the Windows Sound System standard,
//	along with support for Sound Blaster and MIDI emulation.
//
//	NOTE: This procedure reflects the specifications found in Analog Devices'
//		  documentation for the Personal Sound System (the 1848, 2115, etc chipset).
//		  This documentation currently is available via FTP (ftp://ftp.analog.com).
//
InitSw32 (BOOL bForceReset, PCSZ pszFile, WORD wWssBase, WORD wIrq, WORD wDma)
{
    WORD wPortOffset = 0x0;		// offset from 0x220 of SB emultation port (0x0 or 0x20)
    WORD wByte;					// byte read from the load file
    FILE* fIn;					// psuedo-file (for compatibility w/ DOS version)
	WORD wConfigSpec;			// encoding for configuration specification
    int i,j;

    DPF5("InitSw32(%d, %s, %x, %x, %x)", bForceReset, pszFile, wWssBase, wIrq, wDma);

    // Add support for Revision 'B' PSS ASIC
#if TRUE
    DPF1("Checking port signature at 0x%x", PSS_ID_VERS);
    if((InPortWord(PSS_ID_VERS) != (WORD)0x4501) && (InPortWord(PSS_ID_VERS) != (WORD)0x4502)) {
        wPortOffset = (WORD)0x0020;
               
    	DPF1("Checking port signature at 0x%x", PSS_ID_VERS + wPortOffset);
        if((InPortWord(PSS_ID_VERS + wPortOffset) != (WORD)0x4501) 
                && (InPortWord(PSS_ID_VERS + wPortOffset) != (WORD)0x4502)) {  
            DPF( "No SoundWave32/GameWave32 was found!");
            return (FALSE);
            }

        }
	//	Revise the global variable for port address that we will actually use
	uActualSoundBlasterModePort = PSS_DATA + wPortOffset;
#else
	wPortOffset = (WORD)uSoundBlasterModePort - PSS_DATA;
    DPF1("Checking port signature at 0x%x", PSS_ID_VERS + wPortOffset);
    if((InPortWord(PSS_ID_VERS + wPortOffset) != (WORD)0x4501) 
        && (InPortWord(PSS_ID_VERS + wPortOffset) != (WORD)0x4502)) {  
        DPF( "No SoundWave32/GameWave32 was found!");
        return (FALSE);
        }
#endif
   
    DPF("Opening 'file' (table hardcoded into the driver)");
    fIn = OpenBinaryFile(pszFile);

	//	Display what we will be doing (to the kernel debugging stream and/or mono monitor)
	//	NOTE: wIrq and wDma are not specified in any of the port registers
	//		  so these parameters can be removed
    DPF("Setting WSS emulation parameters as follows:");
    DPF3("    WSS at 0x%x; IRQ 0x%x;  DMA = 0x%x", wWssBase, wIrq, wDma);
    DPF2("IRQ Spec is 0x%x; DMA Spec is 0x%x", GetIrqSpec(wIrq), GetDmaSpec(wDma));

	//	Write the encoding for the Windows Sound System base address
    OutPortWord(WSS_CONFIG + wPortOffset, wWssBase << 4);  // eg, set WSS to 530h (5300)

	//	Write the SB emulation IRQ and DMA encoding
	wConfigSpec = GetIrqSpec(uSoundBlasterIRQ) | GetDmaSpec(uSoundBlasterDMA);
    OutPortWord(SB_CONFIG + wPortOffset, wConfigSpec);
    DPF1("SB_CONFIG config set to 0x%x", wConfigSpec)

	//	Write the MIDI emulation port and IRQ encoding
	wConfigSpec = ((WORD)uMidiPort << 4) | GetIrqSpec(uMidiIRQ);
    OutPortWord(MIDI_CONFIG + wPortOffset, wConfigSpec);
    DPF1("MIDI_CONFIG config set to 0x%x", wConfigSpec)

	//	Write the CD interface port and IRQ encoding
	wConfigSpec = ((WORD)uCdInterfacePort << 4) | GetIrqSpec(uCdInterfaceIRQ);
    OutPortWord(CD_CONFIG + wPortOffset, wConfigSpec);
    DPF1("CD_CONFIG config set to 0x%x", wConfigSpec)

	//	Reboot the DSP   
    DPF("Sending reboot");
    wByte = GetNextByte(fIn);                     		// get first byte of boot sequence
    OutPortWord(PSS_DATA + wPortOffset,(WORD)0x00fe);   // send reboot command to existing software
   
    if (bForceReset) {
       i=2048;
       }
    else {
       for(i=0; i<2048; i++) {
           if(InPortWord(PSS_DATA + wPortOffset) == (WORD)0x5555)
               break;
           }
       }
   
    DPF("Sending reset");
    OutPortWord(PSS_DATA + wPortOffset, wByte);          // latch out first byte of boot
    if(i==2048) {
        OutPortWord(PSS_CONTROL + wPortOffset,(WORD)0x2000);   // reset the DSP
        for (i=0; i<32767; i++) {
            ;
            }
        OutPortWord(PSS_CONTROL + wPortOffset,(WORD)0x0000);   // set the DSP reset line high
        DPF("Hard reset performed.");
        }
    else {
        DPF("Hard reset not performed.");
        }
   
    DPF("Loading 'file'");
    for (i=1; i<=3072; i++) {
        for(j=0; j<2048; j++) {
            if(InPortWord(PSS_STATUS + wPortOffset) & PSS_FLAG3) {
                break;
                }
            }
        if(j==2048) {
            DPF("TIMEOUT ERROR after x bytes");
            return (FALSE);
            }
        OutPortWord(PSS_DATA + wPortOffset, GetNextByte(fIn));
        if(i==3072) {
            for(j=0; j<2048; j++) {
                if(InPortWord(PSS_STATUS + wPortOffset) & PSS_FLAG3) {
                    break;
                    }
                }
            if(j==2048) {
                DPF("TIMEOUT ERROR after x bytes");
                return (FALSE);
                }
            wByte = GetNextByte(fIn);
            OutPortWord(PSS_DATA + wPortOffset, wByte);
            i=0;
            }
        if(EndOfFile(fIn)) {
            break;
            }
        }
    CloseBinaryFile(fIn);

    DPF("Setting misc stuff (eg, game support)");
    OutPortWord(PSS_DATA + wPortOffset,0);
	//	Enable the game port
	//	TODO: make this an option
    OutPortWord(PSS_CONFIG + wPortOffset,(WORD)0x0400);
   
   
    while (!(InPortWord(PSS_STATUS + wPortOffset) & PSS_READ_FULL)) {
        ;
        }
    wByte = InPortWord(PSS_DATA + wPortOffset);
    
    DPF1("Version is %x", wByte);
    wByte = InPortWord(PSS_STATUS + wPortOffset);
    DPF1("Status = %x", wByte);
    DPF1("Number of IN's from ports = %d", lNumInPorts);
    DPF1("Number of OUT's to ports = %d", lNumOutPorts);

    DPF("InitSw32: Out");

    return (TRUE);
}
